#!/usr/bin/perl
#
#
#
#---------------------------------------------------------------------------------
# Daniel Segre'
# Harvard Medical School, George Church's lab
# December 3, 2002
# For questions, comments, bugs, please write to:
# dsegre@genetics.med.harvard.edu
#---------------------------------------------------------------------------------
#
#
#
# Following copyright statements to be kept in this program.
# We are using GLPK for linear programming and OOQP for quadratic programming
#
#
#
#  ********************* GLPK ******************************************************
#
# The GLPK package is a part of the GNU project, released under the aegis of GNU. 
#
# Copyright (c) 2000, 2001 Andrew Makhorin, Department for Applied Informatics, Moscow Aviation 
# Institute, Moscow, Russia. All rights reserved. 
#
# Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111, USA. 
#
# Verbatim copying and distribution of this entire article is permitted in any medium, 
# provided this notice is preserved. 
#
#
#  ******************** OOQP *******************************************************
#
# COPYRIGHT 2001 UNIVERSITY OF CHICAGO
#
# The copyright holder hereby grants you royalty-free rights to use, reproduce, prepare 
# derivative works, and to redistribute this software to others, provided that any changes 
# are clearly documented. This software was authored by:
#
#    E. MICHAEL GERTZ      gertz@mcs.anl.gov
#    Mathematics and Computer Science Division
#    Argonne National Laboratory
#    9700 S. Cass Avenue
#    Argonne, IL 60439-4844
#
#    STEPHEN J. WRIGHT     swright@cs.wisc.edu
#    Computer Sciences Department
#    University of Wisconsin
#    1210 West Dayton Street
#    Madison, WI 53706   FAX: (608)262-9777

# Any questions or comments may be directed to one of the authors.

# ARGONNE NATIONAL LABORATORY (ANL), WITH FACILITIES IN THE STATES OF ILLINOIS AND IDAHO, IS 
# OWNED BY THE UNITED STATES GOVERNMENT, AND OPERATED BY THE UNIVERSITY OF CHICAGO UNDER 
# PROVISION OF A CONTRACT WITH THE DEPARTMENT OF ENERGY.
#
#
#
# ********* END of Copyright statements *************************************************

open(OUTFID, ">output") or die "Can't open file: $!\n";

print OUTFID "\n\n";
print OUTFID "******************************************************************\n";
print OUTFID "Program moma_Dec3_2002_win.pl\n";
print OUTFID "by Daniel Segre', dsegre\@genetics.med.harvard.edu\n";
#print scalar localtime;
print OUTFID "\n";
print OUTFID "******************************************************************\n";


# Following definition of main work directory, where executable LP/QP
# are to be found.


# For Peter's Java OAA version, set:
 $homedir="../../../perl";

# For running locally, the setting is:
# $homedir=".";


#-----------------------------------------------------------------------------------
# READ PARAMETERS FILES:
#-----------------------------------------------------------------------------------
print OUTFID "Reading parameter files...\n";
$S_matrix_filename="$homedir/S.par";
$objvector_filename="$homedir/objvector.par";
$constraints_filename="$homedir/constraints.par";

read_S_par($S_matrix_filename);
print OUTFID "S matrix read from $S_matrix_filename \n";

read_constraints_par($constraints_filename);
print OUTFID "Constraints read from $constraints_filename\n";

read_objvector_par($objvector_filename);
print OUTFID "Objective function read from $objvector_filename\n";

print OUTFID "Done.\n";
print OUTFID "N. of fluxes : $N \n";
print OUTFID "N. of metabolites : $M \n";


#-----------------------------------------------------------------------------------
#  READ FILE WITH Isozyme switch and list of knockouts
#
# As of November 20, 2002, the file has the following format:
#
# i         (i=1 means delete isozymes, i=0 means don't)
# vGluMax   (upper bound for glucose uptake)
# vNMax     (upper bound for Nitrogen uptake)
# vO2Max    (upper bound for Oxygen uptake)
# mutant1   (number of gene to be deleted)
# mutant2   (next , etc... as many mutant as one wishes)
# mutant3
# ...
#

$file_with_mutants="mutants.par";
read_mutants($file_with_mutants);

# Modifies constraints according to input from form:
# NOTE: following uptake flux modifications are specific 
#        for flux numbers in Palsson's E.coli stoichiometric matrix.
$constraints{$enzyme[763]} = [ "0" , $GLC_uptake ];
$constraints{$enzyme[923]} = [ "0" , $O2_uptake ];
$constraints{$enzyme[897]} = [ "0" , $N_uptake ];
print OUTFID "Uptake constraints modified according to user input\n";
print OUTFID "Glucose  : $constraints{$enzyme[763]}[1]\n";
print OUTFID "O2       : $constraints{$enzyme[923]}[1]\n";
print OUTFID "Nitrogen : $constraints{$enzyme[897]}[1]\n";


print OUTFID "N. of mutants = $howmanymuts \n";
for ($i = 0  ; $i < $howmanymuts ; $i++) {
  print OUTFID "Mutant n.$i from file: $all_mutant_gene_numbers[$i] \n";
}



#-----------------------------------------------------------------------------------
# BASIC SIMULATION OPTIONS:
#-----------------------------------------------------------------------------------

# QP kind of computation: (NOTE: as of Jan 29, Normalized doesn't really work...)
$normalized_QP=0;   # (0=normal QP ; 1=QP with fluxes normalized with respect to their wt)
if ($normalized_QP==1) {
  print OUTFID "Running QP in NORMALIZED MODE\n";
} else {
  print OUTFID "Running QP in standard mode\n";
}

# NOTE: Following position is specific for Palsson's E.coli parameter file
$Vgr_position=602; # position of growth flux for complete Ecoli
print OUTFID "Postion of Vgro for QP = $Vgr_position\n";

# Boundary to be used in (quasi) complete knock-out:
$ko_threshold_value=0;


$write_fluxes_flag=1;
if ($write_fluxes_flag==1) {
  print OUTFID "Output writing ON\n";
} else {
  print OUTFID "Output writing OFF\n";
}


#----------------------------------------------------------------------------------
# ASSIGN mps FILENAMES 
# These are the files being generated by this Perl script
# and read by the LP/QP solvers
#----------------------------------------------------------------------------------
# Wild type LP:
$LP_mps_wt_filename="Ecoli_complete_wt.mps";

# Knockout QP:
$QP_mps_filename="ecoli_QP_mut.qps";

# Knockout LP:
$LP_mps_filename="ecoli_LP_mut.mps";

if ( -e $LP_mps_wt_filename) {
	system("rm $LP_mps_wt_filename");
}

if ( -e $LP_mps_filename) {
	system("rm $LP_mps_filename");
}

if ( -e $QP_mps_filename) {
	system("rm $QP_mps_filename");
}




#-----------------------------------------------------------------------------------
# INITIALIZE OUTPUT FILES
#-----------------------------------------------------------------------------------
print OUTFID "File initialization...";

#open(FID, ">QP_LP_Vgrowth.dat") or die "Can't open file: $!\n";
#close(FID);

open(FID, ">QP_fluxes.dat") or die "Can't open file: $!\n";
close(FID);

open(FID, ">LP_fluxes.dat") or die "Can't open file: $!\n";
close(FID);

#open(TMPID,">tmp")  or die "Can't open file: $!\n";
#close(TMPID);
#open(FID, ">summary_table.dat") or die "Can't open file: $!\n";
#close(FID);
print OUTFID "Done. \n";


#-----------------------------------------------------------------------------------
# CALCULATE WILD TYPE with FBA
#-----------------------------------------------------------------------------------
print OUTFID "Preliminary part : Calculate wild type with LP \n";
calculate_wildtype();


#-----------------------------------------------------------------------------------
# DO MUTANT
#-----------------------------------------------------------------------------------
# Old part: get mutant number and upper bound from stdin:

# Prepare for LP:
foreach $a (keys %constraints){
  $constraints_LP{$a}=$constraints{$a};
}


# OPTIONAL: choose from stdin whether to read wt from file or use internally computed one: 

# print OUTFID "What starting flux point should be used in QP ? (0 = wild type ; 1 = read from file): ";
# $starting_flux_choice=<STDIN>;
chomp($starting_flux_choice);

$starting_flux_choice=0;

if ($starting_flux_choice==0) {
   @Projectandum_flux=@LP_wt_flux;
} elsif ($starting_flux_choice==1) {

  print OUTFID "Name of file: ";
  $starting_flux_filename=<STDIN>;
  chomp($starting_flux_filename);
  open(FID,"$starting_flux_filename");
  @alllines=<FID>;
  close(FID);
  chomp(@alllines);
  @Projectandum_flux=@alllines;
}



# Prepare for QP:
if ($normalized_QP==1) {  # prepare for normalized QP

  print OUTFID "NORMALIZED QP currently not implemented - see earlier versions\n";

} else {  # ... or prepare for REGULAR QP (not normalized)

  $c=0;
  foreach $flux(@enzyme) {
    #$objhash_QP{$flux}=$LP_wt_flux[$c];
    $objhash_QP{$flux}=$Projectandum_flux[$c];
    $c++;
  }

  foreach $a (keys %constraints){
    $constraints_QP{$a}=$constraints{$a};
  }
}



undef %delenda_fluxa;    # empty hash


# Here below the program look for isoenzymes and for all reactions catalyzed by that gene:

# Old method for input: (receiving number of genes rather than names)
# For name input, as developed by Wayne, see Previous mpa_Jul versions
for ($i_manymuts=1 ; $i_manymuts<=$howmanymuts ; $i_manymuts ++){
  $i_mut=$all_mutant_gene_numbers[($i_manymuts-1)];
  chomp($i_mut);#
  $i_mut_array=$i_mut-1;
  $name_of_mutant=$enzyme[$i_mut_array];


  print OUTFID "Name of mutant: $name_of_mutant \n";
  print OUTFID "Wild type flux of $name_of_mutant: $LP_wt_flux[$i_mut_array] \n";
 
  if ($elimin_isozymes==1) {
    undef %grouped_flux; # deletes content of previous lists of isozymes
    # LOOKS FOR ISOENZYMES OF THE CURRENT ENZYME  -------------------------------------
    # reminder $S[metabolite][enzyme]
    print OUTFID "111111 Searching for isoenzymes of the current enzyme\n";
    $num_alias=$i_mut_array;
    $grouping_counter=0;
    for ($flux_loop=0 ; $flux_loop<$N  ; $flux_loop++){
      $sum_squares=0;
      for ($metab_loop=0 ; $metab_loop<$M ; $metab_loop++){
        $sum_squares=$sum_squares + ( $S[$metab_loop][$flux_loop] != $S[$metab_loop][$num_alias] );
      }
      if ($sum_squares==0) {
        $grouped_flux{$grouping_counter}=$enzyme[$grouping_counter];
        print OUTFID "111111 Isozyme $enzyme[$grouping_counter] added to hash \n";
      }
      $grouping_counter++;
    } # end loop on all fluxes, looking for isoenzymes of current alias

    print OUTFID "111111 Summary of Isozymes:\n";
    for $num_grouped(keys %grouped_flux) {
      print OUTFID "111111 Isoenzyme n.$num_grouped ($grouped_flux{$num_grouped}) added\n";
    }
  } elsif ($elimin_isozymes==0) {
    undef %grouped_flux;
    print OUTFID "000000 NOT searching for isoenzymes...\n";
    $grouped_flux{$i_mut_array}=$name_of_mutant;
  }

    # FOR EACH ISOZYME SEARCH FOR ALL REACTIONS IT CATALYZES (BY NAME SIMILARITY)  
    undef %delenda_fluxa_tmp;    # empty temporary hash
    for $num_grouped(keys %grouped_flux) { # starts loop on all isozymes found before
      $name_of_iso=$enzyme[$num_grouped];
      print OUTFID "222222 Now looking for multiple functions of $name_of_iso...\n";
      $delenda_fluxa_tmp{$num_grouped}=$name_of_iso; # enters in list the isozymne itself
      $elab_mutant=$name_of_iso;
      $elab_mutant =~ s/R$//;    # eliminates final R, if present (reversible)
      if ($elab_mutant =~ /(\D+)(\d+)/) {
        $pure_name=$1;
        $gene_version=$2;
        #print OUTFID "Pure gene name: $pure_name  (Current version: $gene_version)\n";
        # Now searches through all fluxes for those with same enzyme:
        $alias_counter=0;
        for $pot_al_enzyme (@enzyme) {  # loop on potential alias enzymes...
          $pot_al_enz_tmp= $pot_al_enzyme ;
          $pot_al_enz_tmp =~ s/R$//;
          if ($pot_al_enz_tmp =~ /(\D+)(\d+)/) {
            $alias_pure_name=$1;
            $alias_gene_version=$2;
            if ($alias_pure_name eq $pure_name) {
              $delenda_fluxa_tmp{$alias_counter}=$pot_al_enzyme;
              print OUTFID "222222 Flux $pot_al_enzyme added to hash of fluxes to be constrained\n";
            }
          }
          $alias_counter++;
        }
      }
    }
  
   open (FID,">>summary_table.dat");
    print FID "<$i_mut> ";
    print OUTFID "222222 Final report of fluxes being constrained: \n";
    for $num_alias(sort keys %delenda_fluxa_tmp) {
      print FID "$delenda_fluxa_tmp{$num_alias} ";
      print OUTFID "222222  Alias n.$num_alias ($delenda_fluxa_tmp{$num_alias})\n";
    }
    print FID "\n";
    close(FID);
  
  
   # HERE POSSIBLY ALLOW CHOICE OF DOING PARTIAL DELETIONS... (see mpa_Jul31_2002.pl)

    $lowerbound=0;

    $upperbound=0;


  # STARTS LOOP ON ALL FLUXES TO BE DELETED TOGETHER, FOR DEFINING CONSTRAINTS
  for $num_alias(keys %delenda_fluxa_tmp) {
 
    #-----------------------------------------------------------------------------------------
    # Generates Mutant mps file for LP:
    #-----------------------------------------------------------------------------------------

    $constraints_LP{$enzyme[$num_alias]}=[$lowerbound, $upperbound];
    $constraints_QP{$enzyme[$num_alias]}=[$lowerbound, $upperbound];
    print OUTFID "Constraint set for LP and QP: constraint($enzyme[$num_alias])= $lowerbound  $upperbound \n";
  } # end loop on all fluxes to be deleted together

}


data2mps_LP($LP_mps_filename);
do_LP($LP_mps_filename);
($result_LP_mut_status,$LP_obj,@LP_flux)=read_LP_output();

#print OUTFID "flag from LP = $result_LP_mut_status ********************\n\n\n";

if ($result_LP_mut_status==0) {
data2mps_QP($QP_mps_filename);
do_QP($QP_mps_filename);
$QP_obj=0;
($QP_obj,@QP_flux)=read_QP_output();

#  if ($normalized_QP==1) {  # retransform fluxes in absoulte coordinates
#    for ($c=0 ; $c<N ; $c++) {
#      if ($QP_flux[$c] !=0){
#        $QP_flux[$c]=$QP_flux[$c]*$LP_wt_flux[$c];
#      }
#    }
#  }


$LP_Vgr=$LP_flux[$Vgr_position-1];
$QP_Vgr=$QP_flux[$Vgr_position-1];

write_output_to_file();
  
#print  "\n";
#print  "========================================================================\n";
#print  "OPTIMAL MUTANT PREDICTION OUTPUT (LP):\n";
#print  "V_Growth = \t$LP_Vgr\n";
# NOTE: following part is S.par specific for complete E.coli!
#print  "MUTANT (LP) UPTAKE fluxes:\n";    
#print  "\tGLCxtI (764) = \t$LP_flux[763]  \n";
#print  "\tO2xtI (924) = \t$LP_flux[923]  \n";
#print  "\tNH3xtI (898) = \t$LP_flux[897] \n";
#print  "========================================================================\n";


  
#print  "\n";
#print  "========================================================================\n";
#print  "SUBOPTMAL MUTANT PREDICTION OUTPUT (QP):\n";
#print  "V_Growth = \t$QP_Vgr\n";
# NOTE: following part is S.par specific for complete E.coli!
#print  "MUTANT (QP) UPTAKE fluxes:\n";    
#print  "\tGLCxtI (764) = \t$QP_flux[763]  \n";
#print  "\tO2xtI (924) = \t$QP_flux[923]  \n";
#print  "\tNH3xtI (898) = \t$QP_flux[897] \n";
#print  "========================================================================\n";

print OUTFID "\n";

} else {

print OUTFID "\n\n";
print OUTFID "Problem detected: \n";
print OUTFID "The specific constraints imposed for this mutant lead to infeasibility.\n";
print OUTFID "Please report this mutant combination to Daniel Segre' ";
print OUTFID "(dsegre\@genetics.med.harvard.edu). Thank you.\n";

#close(OUTFID);

}
print OUTFID "Closing output file\n";
close (OUTFID);

#----------------------------------------------------------------------------------------
#                            S U B R O U T I N E S
#----------------------------------------------------------------------------------------

sub calculate_wildtype{
# This subroutine  Calculate wild type fluxes (LP)

# Copy original constraints into LP specific constraints
foreach $a (keys %constraints){
  $constraints_LP{$a}=$constraints{$a}; }
data2mps_LP($LP_mps_wt_filename);                # Generate LP mps file
do_LP($LP_mps_wt_filename);                      # Performs optimization
$LP_wt_obj=0;
($result_LP_wt_status,$LP_wt_obj,@LP_wt_flux)=read_LP_output();       # Read output

if ($result_LP_wt_status==0) {

#print  "\n";
#print  "========================================================================\n";
#print  "WILD TYPE PREDICTION OUTPUT:\n";
#print  "V_Growth = \t$LP_wt_obj\n";

# NOTE: following part is S.par specific for complete E.coli!
#print  "Optimized UPTAKE fluxes:\n";
#print  "\tGLCxtI (764) = \t$LP_wt_flux[763]  \n";
#print  "\tO2xtI (924) = \t$LP_wt_flux[923]  \n";
#print  "\tNH3xtI (898) = \t$LP_wt_flux[897] \n";
#print  "========================================================================\n";

#print  "\n";

# Writes wt fluxes to file:
open(FID,">wt_fluxes.dat");
for  ($ii=0; $ii<($N-1) ; $ii++) {
  print FID "$LP_wt_flux[$ii]\n";
}

print FID "$LP_wt_flux[$N-1]";

} else {


#print  "\n";
#print  "========================================================================\n";
#print  "WILD TYPE PREDICTION OUTPUT: INFEASIBLE - Check constraints\n";
#print  "========================================================================\n";

 
}

}

#----------------------------------------------------------------------------------------

sub do_LP {

print OUTFID "Running Linear Programming...";
# EXECUTES Linear Programming using GNU Linear Programming Kit:
system("$homedir/glpsol.exe $_[0] --output dan_output > output_LP_trash");
system("cp dan_output LP_output.txt");
print OUTFID "Done.\n";
}

#----------------------------------------------------------------------------------------

sub do_QP {

print OUTFID "Running Quadratic Programming...";
# Removes previous QP output file"
if ( -e "ecoli_QP_mut.out") {
	system("rm ecoli_QP_mut.out");
}
system("$homedir/qpgen-sparse-gondzio.exe $_[0] > output_QP_trash");
print OUTFID "Done.\n";
}


#----------------------------------------------------------------------------------------

sub read_LP_output {
my (@LP_flux);
#print OUTFID "Reading LP output file...";
open(LP_DATA, "LP_output.txt") or die "Can't open file: $!\n";

@lp_data = <LP_DATA>;
close(LP_DATA);
chomp(@lp_data);

$LP_status=0;

# DETECTS Objective function value:
$counter=0;
$flag=0;
while ($flag==0) {
	$line=@lp_data[$counter];
	if ($line=~/Objective/) {
		@line_pieces=split /\s+/,$line;
		$LP_obj=-$line_pieces[3];
		#print OUTFID "LP obj = $LP_obj\n";
		$status_line=@lp_data[$counter+1];
 		#print OUTFID "XXXXXXX Here is the variable status line:  $status_line\n";
		if ($status_line=~/INFEASIBLE/) {
		  print OUTFID "mutant LP output: INFEASIBLE !!!!";
		$LP_status=-1;
		}
		
		$flag=1;
		
	}
	$counter++;
}


# DETECTS Beginning of Fluxes values:
$flag=0;
while ($flag==0) {
        $line=@lp_data[$counter];
	if ($line=~/Column name/) {
		$flag=1;
	}
	$counter++;
}


# READS LP fluxes

$counter++;
while ($counter<$#lp_data) {
        $line=@lp_data[$counter];
        @line_pieces = split /\s+/, $line;
	$flux=$line_pieces[4];
	push @LP_flux, $flux;
	$counter++;
}

#print OUTFID "Done.\n";
return ($LP_status,$LP_obj,@LP_flux);

}

#----------------------------------------------------------------------------------------

# MODIFIED on Nov 7 2002 for OOQP
sub read_QP_output {
my (@QP_flux);
#print OUTFID "Reading QP output file...";

open(QP_DATA, "ecoli_QP_mut.out") or die "Can't open file: $!\n";

@qp_data = <QP_DATA>;
close(QP_DATA);
chomp(@qp_data);

$counter=8;

# READS FLUXES:
$flag=0;
while ($counter <= $#qp_data) {
#while ($counter <= 30) {
  $line=@qp_data[$counter];
    #print OUTFID "$line\n";
    #$line=@qp_data[$counter];
    @line_pieces = split /\s+/, $line;
    $flux=$line_pieces[3];
    $number=$line_pieces[1];
    #print OUTFID "$number -> $flux \n";
    if (abs($flux) < 1e-9) { $flux=0;};
    push @QP_flux, $flux;	
  $counter++;
}
#print OUTFID "$line\n";

#print OUTFID "Done.\n";
return ($QP_obj,@QP_flux);

}

#----------------------------------------------------------------------------------------

sub write_output_to_file {

#print OUTFID "Writing to file...";
#open(FID, ">>QP_LP_Vgrowth.dat") or die "Can't open file: $!\n";
#print FID "$LP_Vgr\t$QP_Vgr\n";

#print FID "i_mut=$i_mut\t($name_of_mutant)\t\t";
#print FID "wt_flux=$LP_wt_flux[$i_mut_array]\t\t";
#print FID "flux_up_bnd=$upperbound \t\t";
#print FID "Vgr_LP=$LP_Vgr\t\tVgr_QP=$QP_Vgr\n";

#print FID "$LP_Vgr\t\t$QP_Vgr\n";
#close(FID);

if ($write_fluxes_flag==1){
  open(FID, ">>LP_fluxes.dat") or die "Can't open file: $!\n";
  for ($_=0; $_< ($N-1); $_++) {
    print FID "$LP_flux[$_]\n";
  }
  print FID "$LP_flux[$N-1]";
  close(FID);

  open(FID, ">>QP_fluxes.dat") or die "Can't open file: $!\n";
  for ($_=0; $_< ($N-1); $_++) {
    print FID "$QP_flux[$_]\n";
  }
  print FID "$QP_flux[$N-1]";
  close(FID);
};

#print OUTFID "Writing done.\n";
}
#----------------------------------------------------------------------------------------

#****************************************************************************
# FOLLOWING SUBROUTINES ORIGINALLY FROM read_par_files_Jan12.pl
#****************************************************************************

sub read_S_par {

# READING THE STOICHIOMETRIC MATRIX PARAMETER FILE
#---------------------------------------------------------------------

# print OUTFID "Reading file of S matrix...";
open(FID,"$_[0]");
@alllines=<FID>;
close(FID);
chomp(@alllines);

@enzyme = split /\s+/, $alllines[0];
$N=scalar(@enzyme);

shift(@alllines);
$M=scalar(@alllines);

$counter1=0;
foreach $line(@alllines) {
  ($currmetabolite, $rest)= split /\s+/, $line, 2;
  push @metabolite, $currmetabolite;
  @Sline= split /\s+/, $rest;
  for ($counter2=0 ; $counter2<$N ; $counter2++){
    $S[$counter1][$counter2]=$Sline[$counter2];
  }
  $counter1++;
}


# Generate hash of hashes with all fluxes:
for ($i=0 ; $i<$N ; $i++){
  #print OUTFID "$enzyme[$i] ) ";
  for ($j=0 ; $j<$M ; $j++) {
    if ($S[$j][$i] != 0 ) {
      #print OUTFID " $S[$j][$i] $metabolite[$j] ";
      $Shash{$enzyme[$i]}{$metabolite[$j]} = $S[$j][$i];
    }
  }
  #print OUTFID "\n";	
}


# optional: generate matlab readable lists:
if (1==1) {

# Prints list of enzymes into a matlab file:
open(FID,">enzyme.m");
print FID "Enzyme=\{\n";
for $k1 (@enzyme) {
  print FID "\'$k1\'\n";
}
print FID "\};\n";
close(FID);


# Prints list of reactions into a matlab file:
open(FID,">reaction.m");
print FID "Reaction=\{\n";
for $k1 (@enzyme) {
  print FID "\'$k1 ) ";
  for $k2 (keys %{$Shash{$k1}} ) {
   print FID "$Shash{$k1}{$k2} $k2 ";
  }
  print FID "\'\n";
}
print FID "\};\n";
close(FID);

# print OUTFID "Done.\n";

}


}

#****************************************************************************

sub read_constraints_par {

# READING THE CONSTRAINTS PARAMETER FILE
#---------------------------------------------------------------------

# print OUTFID "Reading file of constraints...";
open(FID,"$_[0]");
@alllines=<FID>;
close(FID);
chomp(@alllines);

foreach $line(@alllines) {
($number, $name, $min, $max)= split /\s+/, $line;
$constraints{$name} = [ $min , $max ];
}

#for $_ (keys %constraints) {
#  print OUTFID "$_ : min=$constraints{$_}[0]  -  max=$constraints{$_}[1] \n";
#}

# print OUTFID "Done.\n";

}

#****************************************************************************

sub read_objvector_par {


# READING THE OBJECTIVE FUNCTION PARAMETER FILE
#---------------------------------------------------------------------

# print OUTFID "Reading file of objective function...";
open(FID,"$_[0]");
@alllines=<FID>;
close(FID);
chomp(@alllines);


foreach $line(@alllines) {
  ($name, $value)= split /\s+/, $line;
  $objhash{$name} = $value;
}


#foreach $_(@enzyme) {
#  print OUTFID "$_ -> $objhash{$_}\n";
#}

# print OUTFID "Done.\n";

}


#****************************************************************************

sub read_mutants {

# print OUTFID "Reading file of objective function...";
open(FID,"$_[0]");
@alllines=<FID>;
close(FID);
print OUTFID "@alllines\n";
chomp(@alllines);


# SWITCH FOR DELETION OF ISOZYMES: (0=don't delete ; 1=delete)
$elimin_isozymes=$alllines[0];
print OUTFID "Isozymes deletion switch (0=don't delete ; 1=delete) = $elimin_isozymes\n";

$length_al=$#alllines;

print OUTFID "Total number of lines in par file = $length_al\n";

$GLC_uptake=$alllines[1];
$O2_uptake=$alllines[2];
$N_uptake=$alllines[3];

for ($i = 4  ; $i <= $length_al ; $i++) {
  $all_mutant_gene_numbers[($i-4)]=$alllines[$i];
  print OUTFID "Mutant n.$i from file: $all_mutant_gene_numbers[$4-1] \n";
}

$howmanymuts=$length_al-3;
print OUTFID "N. of mutants = $howmanymuts \n";

}


#****************************************************************************

sub data2mps_LP {

# Generate mps file for LP

#print OUTFID "Entered subroutine data2mps_LP \n";
#print OUTFID "This will generate file $_[0] \n";

open(FID,">$_[0]");
close(FID);

open(FID,">$_[0]");

$mps_name='EColi';
$next_line="NAME          $mps_name";
print FID "$next_line\n";

# ROWS
$next_line="ROWS";
print FID "$next_line\n";

# Objective Function
$next_line="* Objective Function:";
print FID "$next_line\n";
$next_line=" N  COST";
print FID "$next_line\n";

for ($i=0 ; $i<$M ; $i++) {
  $next_line=" E  $metabolite[$i]";
  print FID "$next_line\n";
} 

$next_line="COLUMNS";
print FID "$next_line\n";

for $k1 (@enzyme) {
  $costcoeff=$objhash{$k1};
  if ($costcoeff != "0"){
    $next_line=sprintf "    %-8s  COST      %-5.3g",$k1,-$costcoeff;
    print FID "$next_line\n";
  }
  $ccc=0;
  for $k2 (keys %{$Shash{$k1}} ) {
    $coeff=$Shash{$k1}{$k2};
    if ($ccc==0){
      $next_line=sprintf "    %-8s  %-8s  %-5.3e", $k1,$k2,$coeff;
    } else {
      $next_line=sprintf "    %-8s  %-8s  %-5.3e", $k1,$k2,$coeff; 
    }    
    print FID "$next_line\n";
  }
}

$next_line="RHS";
print FID "$next_line\n";

$next_line="RANGES";
print FID "$next_line\n";

$next_line="BOUNDS";
print FID "$next_line\n";

for $c (sort keys %constraints_LP) {
  $min=$constraints_LP{$c}[0];
  $max=$constraints_LP{$c}[1];
  # print OUTFID "$c :  $min    $max\n";

  if ($min eq $max){
    $next_line=sprintf " FX BND1      %-8s  %-5.3e", $c,$min; 
    print FID "$next_line\n";
  } elsif (($min eq "-Inf")&&($max eq "Inf")) {
    $next_line=sprintf " FR BND1      %-8s", $c; 
    print FID "$next_line\n";
  } elsif (($min eq "-Inf")&&($max ne "Inf")) {  # bound only from above
    $next_line=sprintf " MI BND1      %-8s", $c;
    print FID "$next_line\n";
    $next_line=sprintf " UP BND1      %-8s  %-5.3e", $c,$max;
    print FID "$next_line\n";
  } elsif (($min ne "-Inf")&&($max eq "Inf")) {  # bound only from below
    $next_line=sprintf " LO BND1      %-8s  %-5.3e", $c,$min;
    print FID "$next_line\n";
    $next_line=sprintf " PL BND1      %-8s", $c;
    print FID "$next_line\n";
  } else {
    $next_line=sprintf " LO BND1      %-8s  %-5.3e", $c,$min; 
    print FID "$next_line\n";
    $next_line=sprintf " UP BND1      %-8s  %-5.3e", $c,$max; 
    print FID "$next_line\n";
  }
}

$next_line="ENDATA";
print FID "$next_line\n";

close(FID);

#print OUTFID "Subroutine data2mps_LP completed.\n";

}

#****************************************************************************

sub data2mps_QP {

# Generate mps file for LP
# MODIFIED NOVEMBER 7 2002 FOR GENERATING QP mps
# file compatib;e with OOQP rather than IBM-QP

#print OUTFID "Entered subroutine data2mps_QP \n";

open(FID,">$_[0]");
close(FID);

open(FID,">$_[0]");

$mps_name='EColiQP';
$next_line="NAME          $mps_name";
print FID "$next_line\n";

# ROWS
$next_line="ROWS";
print FID "$next_line\n";

# Objective Function
$next_line="* Objective Function:";
print FID "$next_line\n";
$next_line=" N  COST";
print FID "$next_line\n";

for ($i=0 ; $i<$M ; $i++) {
  $next_line=" E  $metabolite[$i]";
  print FID "$next_line\n";
} 

$next_line="COLUMNS";
print FID "$next_line\n";


for $k1 (@enzyme) {
  $costcoeff=$objhash_QP{$k1};
  if ($costcoeff != "0"){
    $next_line=sprintf "    %-8s  COST      %-5.3e",$k1,-$costcoeff;
    print FID "$next_line\n";
  }
  for $k2 (keys %{$Shash{$k1}} ) {
    $coeff=$Shash{$k1}{$k2};
    $next_line=sprintf "    %-8s  %-8s  %-5.3e", $k1,$k2,$coeff;
    print FID "$next_line\n";
  }
}

$next_line="RHS";
print FID "$next_line\n";

$next_line="RANGES";
print FID "$next_line\n";

$next_line="BOUNDS";
print FID "$next_line\n";

for $c (keys %constraints_QP) {
  $min=$constraints_QP{$c}[0];
  $max=$constraints_QP{$c}[1];

  if ($min eq $max){
    $next_line=sprintf " FX BND1      %-8s  %-5.3e", $c,$min; 
    print FID "$next_line\n";
  } elsif (($min eq "-Inf")&&($max eq "Inf")) {
    $next_line=" FR BND1      $c";
    print FID "$next_line\n";
  } elsif (($min eq "-Inf")&&($max ne "Inf")) {  # bound only from above
    $next_line=sprintf " MI BND1      %-8s", $c;
    print FID "$next_line\n";
    $next_line=sprintf " UP BND1      %-8s  %-5.3e", $c,$max;
    print FID "$next_line\n";
  } elsif (($min ne "-Inf")&&($max eq "Inf")) {  # bound only from below
    $next_line=sprintf " LO BND1      %-8s  %-5.3e", $c,$min;
    print FID "$next_line\n";
    $next_line=sprintf " PL BND1      %-8s", $c;
    print FID "$next_line\n";
  } else {
    $next_line=sprintf " LO BND1      %-8s  %-5.3e", $c,$min; 
    print FID "$next_line\n";
    $next_line=sprintf " UP BND1      %-8s  %-5.3e", $c,$max; 
    print FID "$next_line\n";
  }
}

# Now add quadratic part of objective function:
$next_line="QUADOBJ";
print FID "$next_line\n";

for $k1 (@enzyme) {
  $next_line="                        1.0";
  substr($next_line,4,length($k1))=$k1;
  substr($next_line,14,length($k1))=$k1;
  print FID "$next_line\n";
}

$next_line="ENDATA";
print FID "$next_line\n";

close(FID);

}
